/******************************************
 *      the Beos Reversing Engine         * 
 *              by Zadig                  *
 *                                        *
 * This file contains functions useful    *
 * to obtain "usable" files from objdump. *
 ******************************************/


/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <zem_errors.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "misc.h"
#include "output.h"
#include "chainlist.h"
#include "graph.h"

/***********************************************************
 *  Function: RE_PrintHelp                 File: output.c  *
 ***********************************************************
 * Usage:
 *
 * Print RevEng help on stdout
 ***********************************************************/
int32 RE_PrintHelp(void)
{
   printf("RevEng Error while reading parameters.\n");
   printf("\nusage:\n");
   printf("reveng [options] 'exe filename' 'dest filename'\n\n");
   printf("-split : Creates an 'exe filename' file for disassembly\n");    
   printf("        and an 'exe filename.sym' file for symbols.\n");
   printf("-plt   : Prints plt section (not usefull in most cases).\n");
   printf("-dsym  : Print dynamic symbols references list.\n");
   printf("-nodupsym: Remove duplicate symbols in symbols table(only\n");
   printf("             the first occurence is used).\n");
   printf("-symtab \"filename\": filename is used to inject custom symbols \n");
   printf("                      definitions.\n");
   printf("-ctags : Generate tags to ease navigation in the disassembly\n");
   printf("          with vim.\n");
   printf("-graph : Print a call graph for all functions found.\n");
   printf("-nobl  : Do not print blank lines at each code reference.\n");
   printf("-osort=[fosdf]  : sort lists by offset instead of alphabetically.\n");
   printf("                    - f : functions\n");
   printf("                    - o : objects\n");
   printf("                    - s : strings\n");
   printf("                    - df : dynamic functions\n");
   return(Z_OK);
}

/***********************************************************
 * Function: RE_CreateOutputFile           File: output.c  *
 ***********************************************************
 * Usage:
 * Creates the output file(s).
 *
 * This is the function that is called to create the output
 * files. It calls all other functions that print specific
 * parts of the target:
 * - If split option is enabled then creates 2 files.
 * - Prints RevEng banner.
 * - Prints all strings referenced in the target.
 * - Prints all classes prototypes of the target.
 * - Prints all functions names of the target.
 * - Prints all dynamic functions called by the target.
 * - Prints the call graph of the target.
 * - Finally prints the code of the target.
***********************************************************/
int32 RE_CreateOutputFile(Settings *projectSettings,
                          LIST_Mapping *globalMapping)
{
   int32 i_ret;
   uint8 sz_SymbolsFile[STRING_MAX_SIZE];
   FILE *f_DisasmFile, *f_SymbolsFile;
   FILE *f_DynSymbolsFile = NULL;

   if(globalMapping == NULL)
      return(RE_ERROR_PARAMETER);

   f_DisasmFile = fopen(projectSettings->dest_FileName,"w+");
   if(f_DisasmFile == NULL)
      return(FILE_ERROR_OPEN);

   if(projectSettings->splitOutput == Z_TRUE)
   {
      /* get symbols file name */
      sprintf(sz_SymbolsFile,"%s.sym",projectSettings->dest_FileName);
      f_SymbolsFile = fopen(sz_SymbolsFile,"w+");
      if(f_SymbolsFile == NULL)
         return(FILE_ERROR_OPEN);

      /* get dynamic symbols file name */
      if(projectSettings->printDsym == Z_TRUE)
      {
         sprintf(sz_SymbolsFile,"%s.dsym",projectSettings->dest_FileName);
         f_DynSymbolsFile = fopen(sz_SymbolsFile,"w+");
         if(f_DynSymbolsFile == NULL)
            return(FILE_ERROR_OPEN);
      }
   }
   else
   {
      f_SymbolsFile = f_DisasmFile;
      if(projectSettings->printDsym == Z_TRUE)
         f_DynSymbolsFile = f_DisasmFile;
   }

   i_ret = RE_PrintHeader(f_DisasmFile);
   if(i_ret != Z_OK)
      return(i_ret);

   if(projectSettings->splitOutput == Z_TRUE)
   {
      i_ret = RE_PrintHeader(f_SymbolsFile);
      if(i_ret != Z_OK)
         return(i_ret);
      if(projectSettings->printDsym == Z_TRUE)
      {
         i_ret = RE_PrintHeader(f_DynSymbolsFile);
         if(i_ret != Z_OK)
            return(i_ret);
      }
   }

   if(globalMapping[MAP_STRINGS].list_Map != NULL)
   {
      i_ret = LIST_AlphaSort(&globalMapping[MAP_STRINGS]);
      if(i_ret != Z_OK)
         return(i_ret);
      
      fprintf(f_SymbolsFile,"\n[Strings references]\n"); 
      i_ret = RE_PrintStrings(f_SymbolsFile, projectSettings, 
               &globalMapping[MAP_STRINGS],
               projectSettings->offsetSort & OSORT_STRINGS ? SORT_OFFSET : SORT_ALPHA );
      if(i_ret != Z_OK)
         return(i_ret);
   }
   
   /* print symbols */
   if(globalMapping[MAP_SYMBOLS].list_Map != NULL)
   {
      i_ret = LIST_AlphaSort(&globalMapping[MAP_SYMBOLS]);
      if(i_ret != Z_OK)
         return(i_ret);

      fprintf(f_SymbolsFile,"\n[Objects references]\n"); 
      i_ret = RE_PrintObjects(f_SymbolsFile, projectSettings,
               &globalMapping[MAP_SYMBOLS],
               projectSettings->offsetSort & OSORT_OBJECTS ? SORT_OFFSET : SORT_ALPHA );
      if(i_ret != Z_OK)
         return(i_ret);

      fprintf(f_SymbolsFile,"\n[Classes references]\n");
      i_ret = RE_PrintClasses(f_SymbolsFile, projectSettings, 
               &globalMapping[MAP_SYMBOLS],
               projectSettings->offsetSort & OSORT_CLASSES ? SORT_OFFSET : SORT_ALPHA );
      if(i_ret != Z_OK)
         return(i_ret);

      fprintf(f_SymbolsFile,"\n[Functions references]\n");
      i_ret = RE_PrintFunctions(f_SymbolsFile, &globalMapping[MAP_SYMBOLS],
               projectSettings->offsetSort & OSORT_FUNCTIONS ? SORT_OFFSET : SORT_ALPHA );
      if(i_ret != Z_OK)
         return(i_ret);

   }

   /* print dynamic symbols */
   if(projectSettings->printDsym == Z_TRUE)
   {
      if(globalMapping[MAP_DYNSYM].list_Map != NULL)
      {
         i_ret = LIST_AlphaSort(&globalMapping[MAP_DYNSYM]);
         if(i_ret != Z_OK)
            return(i_ret);

         fprintf(f_DynSymbolsFile,"\n[Dynamic Classes references]\n");
         i_ret = RE_PrintClasses(f_DynSymbolsFile, projectSettings, 
                  &globalMapping[MAP_DYNSYM],
                  projectSettings->offsetSort & OSORT_DCLASSES ? SORT_OFFSET : SORT_ALPHA );
         if(i_ret != Z_OK)
            return(i_ret);

         fprintf(f_DynSymbolsFile,"\n[Dynamic Functions references]\n");
         i_ret = RE_PrintFunctions(f_DynSymbolsFile, &globalMapping[MAP_DYNSYM],
                  projectSettings->offsetSort & OSORT_DFUNCTIONS ? SORT_OFFSET : SORT_ALPHA );
         if(i_ret != Z_OK)
            return(i_ret);
      }
   }

   if( (projectSettings->printGraph == Z_TRUE) && 
       (globalMapping[MAP_SYMBOLS].list_Map != NULL) )
   {
      FILE *f_GraphFile;
      
      sprintf(sz_SymbolsFile,"%s.graph",projectSettings->dest_FileName);
      f_GraphFile = fopen(sz_SymbolsFile,"w+");
      if(f_GraphFile != NULL)
      {
         RE_PrintHeader(f_GraphFile);
         fprintf(f_GraphFile, "\n[Call graph]\n");
         RE_PrintGraph(f_GraphFile, &(globalMapping[MAP_SYMBOLS]));
         fclose(f_GraphFile);
      }
   }

   fprintf(f_DisasmFile, "\n[Code Disassembly]\n");  
   i_ret = RE_PrintCode(f_DisasmFile, projectSettings, globalMapping);
   if(i_ret != Z_OK)
      return(i_ret);

   fclose(f_DisasmFile);
   if(projectSettings->splitOutput == Z_TRUE)
   {
      fclose(f_SymbolsFile);
      if(projectSettings->printDsym == Z_TRUE)
         fclose(f_DynSymbolsFile);
   }

   if(projectSettings->ctags)
   {
      i_ret = RE_CreateTagsFile(projectSettings->dest_FileName, 
                                &globalMapping[MAP_SYMBOLS]);
      if(i_ret != Z_OK)
         return(i_ret);
   }
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintHeader               File: output.c  *
 ***********************************************************
 * Usage:
 *
 * Prints RevEng's banner in a file.
 ***********************************************************/
int32 RE_PrintHeader(FILE *f_OutputFile)
{
   uint8 sz_Header[256];

   DBG_PRINT("Entering RE_PrintHeader...\n"); 

   RE_GetHeader(sz_Header);
   fprintf(f_OutputFile, "%s", sz_Header);  

   DBG_PRINT("End of RE_PrintHeader\n"); 
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintCode                 File: output.c  *
 ***********************************************************
 * Usage:
 *
 *  This function prints the code of the target with
 * sections names, functions names, objects and jumps
 * references.
 ***********************************************************/
int32 RE_PrintCode(FILE *f_OutputFile, Settings *projectSettings, 
                   LIST_Mapping *globalMapping)
{
   FILE *f_AsmFile;
   uint8 sz_CurLine[STRING_MAX_SIZE];
   uint8 b_End, b_ProjDone, b_JmpsDone, b_JumpPlt;
   int32 i, i_ret, i_NbRef;
   uint32 i_CurAddr;
   Mapping *curMap, *curMapRef, *curMapJmps;
   Reference *curRef;
   
   DBG_PRINT("Entering RE_PrintCode...\n");     

   if(globalMapping == NULL)
      return(RE_ERROR_PARAMETER);
   
   f_AsmFile = fopen(projectSettings->asm_SrcFileName,"r");
   if(f_OutputFile == NULL)
   {
      return(FILE_ERROR_OPEN);
   }


   curMap = globalMapping[MAP_SYMBOLS].list_Map;
   curMapJmps = globalMapping[MAP_JUMPS].list_Map;
   curMapRef = globalMapping[MAP_REFS].list_Map;
   b_JumpPlt = Z_FALSE;
   i_ret = (int32)fgets(sz_CurLine, STRING_MAX_SIZE, f_AsmFile);
   while(i_ret != 0)
   {  
      /* looks for definition and references */    
      sscanf(sz_CurLine,"%X",(unsigned int*) &i_CurAddr);

      /* prints if begining of section */
      i = 0;
      while(i < MAX_SECTIONS)
      {
         if(i_CurAddr == projectSettings->sections[i].addr)
         {
            if( (strcmp(projectSettings->sections[i].name, ".plt") == 0) &&
                  (projectSettings->printPlt == Z_FALSE) )
            {
               b_JumpPlt = Z_TRUE;
               break;
            }
            else
            {
               fprintf(f_OutputFile,"\nBegining of section %s...\n",
                       projectSettings->sections[i].name);
               b_JumpPlt = Z_FALSE;
               break;
            }
         }
         i++;
      }

      /* don't print plt section */    
      if(b_JumpPlt == Z_TRUE)
      {
         i_ret = (int32)fgets(sz_CurLine, STRING_MAX_SIZE, f_AsmFile);
         continue;
      }
      b_ProjDone = Z_FALSE;
      b_JmpsDone = Z_FALSE;
      b_End = Z_FALSE;

      while(b_End == Z_FALSE)
      {
         if( (curMap == NULL) ||
               (curMap->addr > i_CurAddr) )
         {
            b_ProjDone = Z_TRUE;
         }
         if( (curMapJmps == NULL) ||
               (curMapJmps->addr > i_CurAddr) )
         {
            b_JmpsDone = Z_TRUE;
         }
         if((b_JmpsDone == Z_TRUE) &&
               (b_ProjDone == Z_TRUE) )
         {
            b_End = Z_TRUE;
            continue;
         }

         if( (b_ProjDone == Z_FALSE) &&
               (curMap->addr == i_CurAddr))
         {
            if(curMap->type == OBJ_TYPE_FUNCTION)
               fprintf(f_OutputFile,"\nFunction %s\n",curMap->label);
            else if(curMap->type == OBJ_TYPE_CLASS)
               fprintf(f_OutputFile,"\nClass %s\n",curMap->label);
            else if(curMap->type == OBJ_TYPE_OBJECT)
               fprintf(f_OutputFile,"\nObject %s\n",curMap->label);

            /* prints where it is referenced */
            i_NbRef = 0;      
            curRef = curMap->references.list_Ref;
            while(curRef != NULL)
            {
               if(i_NbRef == 0)
               {
                  if(curMap->type == OBJ_TYPE_INSTRUCTION)
                  {

                     fprintf(f_OutputFile,
                        "%sReferenced by (conditionnal) jump(s) at Address(es):\n", 
                        projectSettings->blank_line);
                  }
                  else
                  {
                     fprintf(f_OutputFile,"%sReferenced at Address(es):\n",
                        projectSettings->blank_line);
                  }
                  fprintf(f_OutputFile,"\t%.8X  ",(unsigned int)curRef->addr);
               }
               else
               {
                  fprintf(f_OutputFile,"%.8X  ",(unsigned int)curRef->addr);
               }

               i_NbRef++;
               if( ((i_NbRef % 5) == 0) && (i_NbRef != 0) )
               {
                  fprintf(f_OutputFile,"\n\t");          
               }
               curRef = curRef->next;          
            }
            if(i_NbRef != 0)
            {
               fprintf(f_OutputFile,"\n");
            }

         } /* end if(curMap->addr == i_CurAddr) */        

         if( (b_JmpsDone == Z_FALSE) &&
               (curMapJmps->addr == i_CurAddr))
         {
            /* prints where it is referenced */
            i_NbRef = 0;      
            curRef = curMapJmps->references.list_Ref;
            while(curRef != NULL)
            {
               if(i_NbRef == 0)
               {
                  fprintf(f_OutputFile,
                     "%sReferenced by (conditionnal) jump(s) at Address(es):\n",
                        projectSettings->blank_line);
                  fprintf(f_OutputFile,"\t%.8X  ",(unsigned int)curRef->addr);
               }
               else
               {
                  fprintf(f_OutputFile,"%.8X  ",(unsigned int)curRef->addr);
               }

               i_NbRef++;
               if( ((i_NbRef % 5) == 0) && (i_NbRef != 0) )
               {
                  fprintf(f_OutputFile,"\n\t");          
               }
               curRef = curRef->next;          
            }
            if(i_NbRef != 0)
            {
               fprintf(f_OutputFile,"\n");
            }
         } /* end if(curMapJmps->addr == i_CurAddr) */        

         if(b_ProjDone == Z_FALSE)
         {    
            curMap = curMap->next;        
         }
         if(b_JmpsDone == Z_FALSE) 
         {    
            curMapJmps = curMapJmps->next;        
         }
      } /* end while(curMap) */

      while( (curMapRef != NULL) && 
             (curMapRef->addr <= i_CurAddr) )
      {
         if(curMapRef->addr == i_CurAddr) 
         {
            if( (curMapRef->type != OBJ_TYPE_VAR_REFERENCE) &&
                (curMapRef->type != OBJ_TYPE_MESSAGE) )
            {
               fprintf(f_OutputFile,"%s\n",curMapRef->label);
            }
            else               
               break;

            curMapRef = curMapRef->next;
         }
      }

      sz_CurLine[strlen(sz_CurLine)-1] = '\0';  /* remove the '\n' */
      if( (curMapRef != NULL) &&
          (curMapRef->addr == i_CurAddr) )
      {
         if( (curMapRef->type == OBJ_TYPE_VAR_REFERENCE) ||
             (curMapRef->type == OBJ_TYPE_MESSAGE) )
         {
            /* align comments on a 5 char boundary */
            if( ((strlen(sz_CurLine) + 1) % RE_COMMENT_ALIGN != 0) ||
                ((strlen(sz_CurLine) + 1) < RE_COMMENT_MIN)
              )
            {
               uint8 i_PadSize;
               uint8 *psz_String;

               if(strlen(sz_CurLine) + 1 < RE_COMMENT_MIN)
               {
                  i_PadSize = RE_COMMENT_MIN - (strlen(sz_CurLine) + 1);
               }
               else
               {
                  i_PadSize = RE_COMMENT_ALIGN - 
                              ( (strlen(sz_CurLine) + 1) % RE_COMMENT_ALIGN);
               }
               psz_String = sz_CurLine + strlen(sz_CurLine);
               for(i=0; i< i_PadSize; i++)
               {
                  *psz_String = ' ';
                  psz_String++;
               }
               *psz_String = '\0';
            }
            
            /* and print it */
            fprintf(f_OutputFile,"%s%s\n",sz_CurLine, curMapRef->label);
            curMapRef = curMapRef->next;
         }
         else
            fprintf(f_OutputFile,"%s\n",sz_CurLine);
      }
      else
         fprintf(f_OutputFile,"%s\n",sz_CurLine);

      i_ret = (int32)fgets(sz_CurLine, STRING_MAX_SIZE, f_AsmFile);
   }

   fclose(f_AsmFile);
   DBG_PRINT("End of RE_PrintCode\n");     
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintStrings              File: output.c  *
 ***********************************************************
 * Usage:
 *
 *  This function prints all strings referenced in the
 * target. Strings are printed in alphabetical order.
 ***********************************************************/
int32 RE_PrintStrings(FILE *f_OutputFile, Settings *projectSettings,
                      LIST_Mapping *stringsMapping,
                      RE_SortType e_SortType)
{
   Mapping *p_CurMap;

   DBG_PRINT("Entering RE_PrintStrings...\n");     
   
   if(stringsMapping == NULL)
      return(RE_ERROR_PARAMETER);

   if(e_SortType == SORT_ALPHA)
      p_CurMap = stringsMapping->alpha;
   else
      p_CurMap = stringsMapping->list_Map;

   while(p_CurMap != NULL) 
   {
      fprintf(f_OutputFile,"%.8X: %s\n", (unsigned int)p_CurMap->addr, p_CurMap->label);
      if(e_SortType == SORT_ALPHA)
         p_CurMap = p_CurMap->next_alpha;
      else
         p_CurMap = p_CurMap->next;
   }

   DBG_PRINT("End of RE_PrintStrings\n");      
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintObjects              File: output.c  *
 ***********************************************************
 * Usage:
 *
 *  This function prints all objects defined in the target.
 *  Objects are printed in alphabetical order.
 ***********************************************************/
int32 RE_PrintObjects(FILE *f_OutputFile, Settings *projectSettings,
                      LIST_Mapping *symbolsMapping,
                      RE_SortType e_SortType)
{
   Mapping *p_CurMap;

   DBG_PRINT("Entering RE_PrintObjects...\n");     

   if(symbolsMapping == NULL)
      return(RE_ERROR_PARAMETER);

   if(e_SortType == SORT_ALPHA)
      p_CurMap = symbolsMapping->alpha;
   else
      p_CurMap = symbolsMapping->list_Map;

   while(p_CurMap != NULL) 
   {
      if( (p_CurMap->type == OBJ_TYPE_OBJECT) &&            /* don't print rtti symbols */
          (strstr(p_CurMap->label, RTTI_SYMBOL) == NULL) )
         fprintf(f_OutputFile,"%.8X %.8X: %s\n", 
                  (unsigned int)p_CurMap->addr, 
                  (unsigned int)p_CurMap->size, 
                  p_CurMap->label);
      
      if(e_SortType == SORT_ALPHA)
         p_CurMap = p_CurMap->next_alpha;
      else
         p_CurMap = p_CurMap->next;
   }

   DBG_PRINT("End of RE_PrintObjects\n");      
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintClasses              File: output.c  *
 ***********************************************************
 * Usage:
 * This function prints all classes in alphabetical order,
 * and all methods of each class in alphabetical order.
 ***********************************************************/
int32 RE_PrintClasses(FILE *f_OutputFile, Settings *projectSettings,
                      LIST_Mapping *projectMapping,
                      RE_SortType e_SortType)
{
   Mapping  *p_CurClass;
   uint8    sz_ClassName[STRING_MAX_SIZE], *psz_ClassSep, *psz_Space;
   uint8    b_End;

   DBG_PRINT("Entering RE_PrintClasses...\n");     
   
   if(projectMapping == NULL)
      return(RE_ERROR_PARAMETER);

   if(e_SortType == SORT_ALPHA)
      p_CurClass = projectMapping->alpha;
   else
      p_CurClass = projectMapping->list_Map;
   
   while(p_CurClass != NULL)
   {    
      psz_ClassSep = strstr(p_CurClass->label, "::");
      psz_Space = strstr(p_CurClass->label, " ");
      if( (psz_ClassSep == NULL) ||
          (p_CurClass->type != OBJ_TYPE_FUNCTION) ||
          ( (psz_Space != NULL) && (psz_Space < psz_ClassSep)) /* space in the name == not a class */
        )
      {
         if(e_SortType == SORT_ALPHA)
            p_CurClass = p_CurClass->next_alpha;
         else
            p_CurClass = p_CurClass->next;
         continue;
      }
      else
      {
         strcpy(sz_ClassName, p_CurClass->label);
         sz_ClassName[psz_ClassSep - p_CurClass->label] = '\0';
      }
      
      fprintf(f_OutputFile,"\nClass %s\n", sz_ClassName);
      
      /* look for all methods of this class */
      b_End = Z_FALSE;
      while(b_End == Z_FALSE)
      {
         /* check that the symbol is a class method of our class */
         psz_ClassSep = strstr(p_CurClass->label, "::");
         if( (psz_ClassSep == NULL) ||
             ( (psz_ClassSep != NULL) && 
                strncmp(p_CurClass->label, sz_ClassName, 
                        psz_ClassSep - p_CurClass->label) != 0)
           )
         {
            b_End = Z_TRUE;
         }
         else
         {
            if(p_CurClass->type == OBJ_TYPE_FUNCTION)  /* methods virtual tables
                                                          are objects            */
            {
               fprintf(f_OutputFile,"%.8X: %s\n", (unsigned int)p_CurClass->addr, 
                       p_CurClass->label);
            }

            if(e_SortType == SORT_ALPHA)
               p_CurClass = p_CurClass->next_alpha;
            else
               p_CurClass = p_CurClass->next;
            if(p_CurClass == NULL)
               b_End = Z_TRUE;
         }
      } 
   }

   DBG_PRINT("End of RE_PrintClasses\n");       
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_PrintFunctions            File: output.c  *
 ***********************************************************
 * Usage:
 * This function prints all functions name of the target
 * in a file in alphabetical order.
 ***********************************************************/
int32 RE_PrintFunctions(FILE *f_OutputFile, LIST_Mapping *projectMapping,
                        RE_SortType e_SortType)
{
   Mapping *p_CurFunction;
   uint8 *psz_FunctionName, *psz_ClassSep, *psz_Space;

   DBG_PRINT("Entering RE_PrintFunctions...\n");     
   
   if(projectMapping == NULL)
      return(RE_ERROR_PARAMETER);
   
   if(e_SortType == SORT_ALPHA)
      p_CurFunction = projectMapping->alpha;
   else
      p_CurFunction = projectMapping->list_Map;
   while(p_CurFunction != NULL)
   {
      /* don't process templates */
      psz_FunctionName = strstr(p_CurFunction->label,"<"); 
      
      if( (psz_FunctionName == NULL) &&                  /* templates contain '<' and '>' */
          (p_CurFunction->type == OBJ_TYPE_FUNCTION) &&  /* process only functions */
          (strstr(p_CurFunction->label, RTTI_SYMBOL) == NULL) 
        )
      {
         /* check that it is not a class */
         psz_ClassSep = strstr(p_CurFunction->label, "::");
         psz_Space = strstr(p_CurFunction->label, " ");
         if( (psz_ClassSep == NULL) ||
             ( (psz_Space != NULL) && (psz_Space < psz_ClassSep)) /* space in the name == not a class */
           )
         {         
            fprintf(f_OutputFile,"%.8X: %s\n", (unsigned int)p_CurFunction->addr,  
                    p_CurFunction->label);
         }
      }
      if(e_SortType == SORT_ALPHA)
         p_CurFunction = p_CurFunction->next_alpha;
      else
         p_CurFunction = p_CurFunction->next;
   }

   DBG_PRINT("End of RE_PrintFunctions\n");
   return(Z_OK);
}


/***********************************************************
 *  Function: RE_CreateTagsFile            File: output.c  *
 ***********************************************************
 * Usage:
 * This function creates a tags file containing all
 * functions and classes of the target.
 * This will be used by VIM (and other editors) to step
 * inside functions easily.
 ***********************************************************/
#if 1
int32 RE_CreateTagsFile(uint8 *psz_DisasmFileName, 
                        LIST_Mapping *projectMapping)
{
   Mapping *p_CurMap;
   uint8 *psz_NameEnd, *psz_NameStart;
   uint8  sz_TagName[STRING_MAX_SIZE];
   FILE  *f_TagsFile;

   DBG_PRINT("Entering RE_CreateTagsFile...\n");     
   
   /* creates the file */
   f_TagsFile = fopen("tags","w+");
   if(f_TagsFile == NULL)
      return(FILE_ERROR_OPEN);

   /* prints header */
   fprintf(f_TagsFile, CTAG_HEADER, RE_VERSION);

   /* prints functions */
   p_CurMap = projectMapping->alpha;
   while(p_CurMap != NULL)
   {
      /* looks for new class */
      /* don't process templates */
      if( (strstr(p_CurMap->label,"<") == NULL) && /* templates contain '<' and '>' */
          (strstr(p_CurMap->label, RTTI_SYMBOL) == NULL))    /* no rtti symbols */
      {
         /* gets tag name */
         /* for classes, the tag name is the method name */
         /* without params                               */
         psz_NameStart = strstr(p_CurMap->label, "::");
         if(psz_NameStart != NULL)  /* C++ Method */
            psz_NameStart += 2;
         else                       /* C function */
            psz_NameStart = p_CurMap->label;
         
         psz_NameEnd = strstr(psz_NameStart, "(");
         if(psz_NameEnd == NULL)                    /* no params */
         {
            if(psz_NameStart != p_CurMap->label)    /* class is not possible */
            {
               p_CurMap = p_CurMap->next_alpha;
               continue;
            }
            else
            {
               psz_NameEnd = p_CurMap->label + strlen(p_CurMap->label);
            }
         }
         
         strncpy(sz_TagName, psz_NameStart, 
               psz_NameEnd - psz_NameStart);
         sz_TagName[psz_NameEnd - psz_NameStart] = '\0';
         
         /* creates tag entry in file */
         if(p_CurMap->type == OBJ_TYPE_FUNCTION) 
         {
            fprintf(f_TagsFile,"%s\t%s\t/^Function %s$/\n", sz_TagName,  
                    psz_DisasmFileName, p_CurMap->label);
         }
         else if(p_CurMap->type == OBJ_TYPE_CLASS)
         {
            fprintf(f_TagsFile,"%s\t%s\t/^Class %s$/\n", sz_TagName,
                    psz_DisasmFileName, p_CurMap->label);
         }
      }
      p_CurMap = p_CurMap->next_alpha;
   }

   fclose(f_TagsFile);
   DBG_PRINT("End of RE_CreateTagsFile\n");       
   return(Z_OK);
}
#else
int32 RE_CreateTagsFile(uint8 *psz_DisasmFileName, 
                        LIST_Mapping *projectMapping)
{
   Mapping *p_CurMap;
   uint8 *psz_NameEnd, *psz_NameStart;
   uint8  sz_TagName[STRING_MAX_SIZE];
   FILE  *f_TagsFile;

   DBG_PRINT("Entering RE_CreateTagsFile...\n");     
   
   /* creates the file */
   f_TagsFile = fopen("tags","w+");
   if(f_TagsFile == NULL)
      return(FILE_ERROR_OPEN);

   /* prints header */
   fprintf(f_TagsFile, CTAG_HEADER, RE_VERSION);

   /* prints functions */
   p_CurMap = projectMapping->list_Map;
   while(p_CurMap != NULL)
   {
      /* looks for new class */
      /* don't process templates */
      if( (strstr(p_CurMap->label,"<") == NULL) && /* templates contain '<' and '>' */
          (strstr(p_CurMap->label,"global ") == NULL) &&    /* these are rtti ...   */ 
          (strstr(p_CurMap->label," function") == NULL) )   /* ...symbols           */
      {
         /* gets tag name */
         /* for classes, the tag name is the class name */
         /* without params                               */
         psz_NameStart = strstr(p_CurMap->label, "::");
         if(psz_NameStart != NULL)  /* C++ Method */
         {
            psz_NameEnd = strstr(psz_NameStart, "("); /* this is not possible */
            if(psz_NameEnd == NULL)
            {
               p_CurMap = p_CurMap->next;
               continue;
            }

            strncpy(sz_TagName, p_CurMap->label, psz_NameStart - p_CurMap->label); 
            sz_TagName[psz_NameStart - p_CurMap->label] = '\0';

            /* creates tag entry in file */
            if(p_CurMap->type == OBJ_TYPE_FUNCTION) 
            {
               fprintf(f_TagsFile,"%s\t%s\t/^%s$/;\"\tf\tclass:%s\n", sz_TagName,  
                     psz_DisasmFileName, p_CurMap->label, sz_TagName);
            }
         }
         else                       /* C function */
         {
            psz_NameStart = p_CurMap->label;
            
            psz_NameEnd = strstr(psz_NameStart, "(");
            if(psz_NameEnd == NULL)                    /* no params in the funcion  */
               psz_NameEnd = p_CurMap->label + strlen(p_CurMap->label);

            strncpy(sz_TagName, psz_NameStart, 
                  psz_NameEnd - psz_NameStart);
            sz_TagName[psz_NameEnd - psz_NameStart] = '\0';

            /* creates tag entry in file */
            if(p_CurMap->type == OBJ_TYPE_FUNCTION) 
            {
               fprintf(f_TagsFile,"%s\t%s\t/^%s$/;\"\tf\n", sz_TagName,  
                     psz_DisasmFileName, p_CurMap->label);
            }
         }
         
      }
      p_CurMap = p_CurMap->next;
   }

   fclose(f_TagsFile);
   DBG_PRINT("End of RE_CreateTagsFile\n");       
   return(Z_OK);
}

#endif
